"""
    impersonation
"""

import logging

import zoom
import zoom.html as h

logger = logging.getLogger(__name__)

def impersonate(username):
    """Impersonate a user"""
    zoom.authorize()
    user = zoom.get_site().users.locate(username)
    if user:
        session = zoom.system.request.session
        session.impersonated_user = user.username
        zoom.audit('start impersonating', zoom.system.request.user.username, username)
        return user


def get_current_username(request):
    """get actual current user username"""
    site = request.site
    return (
        getattr(request, 'username', None) or
        site.config.get('users', 'override', '') or
        getattr(request.session, 'username', None) or
        request.remote_user or
        site.guest or
        None
    )


def stop_impersonating(request):
    """Stop impersonating"""
    session = zoom.system.request.session
    if hasattr(session, 'impersonated_user'):
        impersonated_user = session.impersonated_user
        del session.impersonated_user
        actual_username = get_current_username(request)
        actual_user = zoom.system.site.users.first(username=actual_username)
        if actual_user:
            zoom.audit(
                'stop impersonating',
                actual_username,
                impersonated_user,
                user=actual_user,
                app_name='middlware'
            )
        else:
            logger.warning('Unable to log stop impersonating - unknown current user')


def get_impersonated_username():
    """Returns username of impersonated user or None"""
    session = zoom.system.request.session
    return getattr(session, 'impersonated_user', None)


def get_impersonation_notice():
    username = get_impersonated_username()
    if username is None:
        return ''
    return str(
        zoom.Component(
            h.div(
                'Impersonating {}'.format(username),
                zoom.link_to(
                    'Stop Impersonating',
                    '/stop-impersonation',
                    classed='action'
                ),
                id='impersonation-notice'
            )
        )
    )


def handler(request, handle, *rest):
    """impersonation handler"""
    if request.path == '/stop-impersonation':
        stop_impersonating(request)
        return zoom.redirect_to('/').render(request)
    return handle(request, *rest)



"""
    performance profiler
"""

from decimal import Decimal
import io
import timeit
import logging
import cProfile
import os
import sys

from zoom.utils import ItemList
from zoom.response import HTMLResponse


def round(value):
    """Round a decimal value

    >>> round(Decimal(1.23422))
    Decimal('1.234')
    """
    return Decimal(value).quantize(Decimal('1.000'))


class SystemTimer(object):
    """time system events

    >>> timer = SystemTimer()
    >>> timer.add('got here')
    >>> timer.add('got there')

    >>> import time
    >>> timer.time('slow step', time.sleep, 0.1)

    >>> report = timer.report()
    >>> len(report.splitlines())
    7

    >>> len(timer.data)
    5
    """

    def __init__(self, start_time=None):
        self.start_time = start_time or timeit.default_timer()
        self.previous_time = self.start_time
        self.record = []
        self.add('modules loaded')

    def add(self, comment):
        """add a measure to the system timer log"""
        current_time = timeit.default_timer()
        entry = (
            comment,
            # '.' * (40 - len(comment)),
            round((current_time - self.previous_time) * 1000),
            round((current_time - self.start_time) * 1000),
        )
        self.record.append(entry)
        self.previous_time = current_time

    def time(self, name, function, *args, **kwargs):
        self.add('starting {}'.format(name))
        result = function(*args, **kwargs)
        self.add('finished {}'.format(name))
        return result

    def report(self):
        """print a report of the timed events"""
        return str(ItemList(self.data, labels=self.labels))

    @property
    def data(self):
        return self.record

    @property
    def labels(self):
        return 'Milestone', 'Time', 'Total'


def get_profile_data(profiler):
    """Capture the stdout printout of the code profiler

    >>> class Profiler(object):
    ...     def print_stats(self):
    ...         print('the stats!')
    >>> get_profile_data(Profiler())
    'the stats!\\n'
    """
    save_stdout = sys.stdout
    try:
        sys.stdout = io.StringIO()
        profiler.print_stats()
        result = sys.stdout.getvalue()
    finally:
        sys.stdout = save_stdout
    return result


def profiled(request, next_handler, *rest):

    def site_profiling(request):
        return hasattr(request, 'user') and hasattr(request, 'site') and request.site.profiling

    def send(message):
        """
        send

        Send the profile data to the page for rendering.
        """

        topic = 'system.debug.%s' % request.user.user_id
        queue = request.site.queues.topic(topic)
        queue.clear()
        queue.send(message)
        logger.debug('sent profile message to %s', topic)

    def is_profilable(result):
        """
        is_profilable

        We only profile regular HTML results.  Not individual images, JSON,
        or other result types.
        """

        # Retrieve headers from the result if it has that property.
        headers = getattr(result, 'headers', dict())
        if not isinstance(headers, dict):
            # Handle the case where the response class is doing something
            # funky with its headers object.
            logger.debug("headers isn't a dict, won't profile")
            headers = dict()

        # Transform headers to lower case because casing isn't invariant.
        headers = {k.lower(): v for k, v in headers.items()}

        # Profile if this is an HTML response.  This is not perfect but it is
        # a reasonable compromise until repsonses are more consistent in expressing
        # their content type.  Looking in the cookies could also be added but
        # it would be better if all responses were just explicit.
        ct = getattr(result, 'content_type', headers.get('content-type', None))
        return ct and 'html' in ct


    profiling_code = request.env.get('ZOOM_CODE_PROFILER')

    logger = logging.getLogger(__name__)

    if profiling_code:
        code_profiler = cProfile.Profile()
        code_profiler.enable()

    result = None
    try:

        result = next_handler(request, *rest)

    finally:

        # We wait until after the request is done to test this
        # because the profiler is installed early enough in the
        # stack that the site is not instantiated until we are
        # on the way back.

        if not site_profiling(request):
            logger.debug('profiler data ignored')
            if not hasattr(request, 'user'):
                logger.debug('profile user missing')
            if not hasattr(request, 'site'):
                logger.debug('profile site missing')
            elif not request.site.profiling:
                logger.debug('profiling turned off')

        else:

            if profiling_code:
                code_profiler.disable()
                code_profile = get_profile_data(code_profiler)
            else:
                code_profile = 'code profiler is off'

            request.profiler.add('finished')

            # TODO: move this to a debug layer that sends other things to the debugger
            message = dict(
                profiler_path=request.path,
                system_profile=request.profiler.record,
                database_profile=['no database'],
                code_profile=code_profile,
            )
            if hasattr(request, 'site'):
                if hasattr(request.site, 'db'):
                    message['database_profile'] = [(
                        round(time * 1000),
                        statement,
                        values,
                        source,
                    ) for time, statement, values, source in request.site.db.get_stats()]

            if is_profilable(result):
                send(message)
            else:
                logger.debug('ignoring profile of response type %s', type(result))

        del request.profiler
    return result


def handler(request, next_handler, *rest):
    """Handle profiled requests"""

    if os.environ.get('ZOOM_PROFILER'):
        request.profiling = True
        return profiled(request, next_handler, *rest)
    else:
        return next_handler(request, *rest)



"""
    admin model
"""

import datetime
import logging
import uuid

import zoom
import zoom.mail
from zoom.context import context
from zoom.tools import today, now
from zoom.users import Users
from zoom.models import Groups


def get_user_group_options(site):
    """return options for a groups pulldown field"""
    user_groups = list(sorted(
        (group.link, group.key)
        for group in site.groups.find(**{'type': 'U'})
    ))
    return user_groups


def get_subgroups(db, groups):
    """get subgroups for a list of groups
    """

    def get_memberships(group, memberships, depth=0):
        """get group memberships"""
        result = set([group])
        if depth < 10:
            for grp, sgrp in memberships:
                if group == sgrp and grp not in result:
                    result |= get_memberships(grp, memberships, depth+1)
        return result

    all_subgroups = list(db(
        'SELECT group_id, subgroup_id FROM subgroups ORDER BY subgroup_id'
    ))

    memberships = set([])
    for group in groups:
        memberships |= get_memberships(group, all_subgroups)

    return memberships


def get_supergroups(db, groups):
    """get supergroups for a list of groups
    """

    def get_memberships(group, memberships, depth=0):
        """get group memberships"""
        result = set([group])
        if depth < 10:
            for grp, sgrp in memberships:
                if group == grp and sgrp not in result:
                    result |= get_memberships(sgrp, memberships, depth+1)
        return result

    all_subgroups = list(db(
        'SELECT group_id, subgroup_id FROM subgroups ORDER BY subgroup_id'
    ))

    memberships = set([])
    for group in groups:
        memberships |= get_memberships(group, all_subgroups)

    return memberships


def named_groups(db, group_ids):
    """Returns names for a list of group ids"""
    result = []
    for _id, name in db('SELECT id, name FROM `groups`'):
        if _id in group_ids:
            result.append(name)
    return result


def get_subgroup_options(db, group_id):
    cmd = """
    select name, id
    from `groups`
    where id <> %s and left(groups.name, 2) <> 'a_'
    """
    return list(
        (name, str(id))
        for name, id in db(cmd, group_id)
    )


def get_role_options(db, group_id):
    cmd = """
    select name, id
    from `groups`
    where id <> %s and left(groups.name, 2) <> 'a_'
    """
    return list(
        (name, str(id))
        for name, id in db(cmd, group_id)
    )


class AdminModel(object):

    def __init__(self, db):
        self.db = db
        self.logger = logging.getLogger(__name__)
        self.groups = Groups(db)

    def log(self, message, *args):
        self.logger.debug(message, *args)

    def get_user_options(self):
        return sorted(
            (user.link, user.user_id) for user in Users(self.db)
        )

    def get_subgroup_options(self, group_id):
        return sorted(
            (group.link, group.group_id)
            for group in self.groups.find(**{'type': 'U'})
            if group.group_id != group_id
        )

    def get_role_options(self, group_id):
        return sorted(
            (group.link, group.group_id)
            for group in self.groups.find(**{'type': 'U'})
            if group.group_id != group_id
        )

    def get_app_options(self):
        groups_lookup = {
            group.name: group.group_id
            for group in zoom.system.site.groups.find(type='A')
        }
        return sorted([
            (
                app.title,
                groups_lookup.get('a_' + app.name, 'a_' + app.name)
            ) for app in zoom.system.site.apps
        ], key=repr)

    def update_group_apps(self, record):
        """Post updated group apps"""

        # create groups for uninstalled apps
        new_groups_required = set(
            app_key for app_key in record['apps']
            if isinstance(app_key, str) and app_key.startswith('a_')
        )
        for group_name in new_groups_required:
            app_name = group_name[2:]
            zoom.system.site.groups.add_app(app_name)

        groups_lookup = {
            group.name: group.group_id
            for group in zoom.system.site.groups.find(type='A')
        }
        new_keys = set(
            groups_lookup.get(app_key, app_key)
            for app_key in record['apps']
        )

        group = record
        group.update_apps_by_id(new_keys)

    def update_group_relationships(self, record):
        group = record
        group.update_members_by_id(group['users'])
        group.update_subgroups_by_id(group['subgroups'])
        group.update_roles_by_id(group['roles'])
        self.update_group_apps(record)


def get_index_metrics(db):

    def count(where, *args):
        """Return the result of a count query"""
        return '{:,}'.format((list(db('select count(*) from ' + where, *args))[0][0]))

    def avg(metric, where, *args):
        """Return the result of a query that calculates an average"""
        value = list(db('select avg({}) from {}'.format(metric, where), *args))[0][0]
        return '{:,.1f}'.format(value) if value else '-'

    the_day = today()
    host = zoom.system.request.host
    recently = now() - datetime.timedelta(minutes=60)

    online_users = count('users where status="A" and last_seen >= %s and username not in ("guest")', recently)
    num_users = count('users where status="A"')
    num_groups = count('groups where type="U"')
    num_requests = count('log where status="C" and server=%s and timestamp>=%s', host, the_day)
    num_errors = count('log where status="E" and server=%s and timestamp>=%s', host, the_day)
    avg_speed = avg('elapsed', 'log where status="C" and server=%s and timestamp>=%s and path<>"/login"', host, the_day)
    num_authorizations = count('audit_log where timestamp>=%s', the_day)

    metrics = [
        (f'Users online of {num_users} active users', '/admin/users', online_users),
        ('Groups', '/admin/groups', num_groups),
        ('Requests Today', '/admin/requests', num_requests),
        ('Errors Today', '/admin/errors', num_errors),
        ('Performance (ms)', '/admin/requests', avg_speed),
        ('Authorizations Today', '/admin/audit', num_authorizations)
    ]
    return metrics


def update_user_groups(record):
    logger = logging.getLogger(__name__)
    db = context.site.db
    record_id = record['_id']

    existing_groups = set(
        group_id for group_id, in
        db('select group_id from members where user_id=%s', record_id)
    )
    logger.debug('existing_groups: %r', existing_groups)

    updated_groups = set(record['memberships'])
    logger.debug('updated_groups: %r', updated_groups)

    if updated_groups != existing_groups:
        if existing_groups - updated_groups:
            logger.debug('deleting: %r', existing_groups - updated_groups)
            cmd = 'delete from members where user_id=%s and group_id in %s'
            db(cmd, record_id, existing_groups - updated_groups)
        if updated_groups - existing_groups:
            logger.debug('inserting: %r', updated_groups - existing_groups)
            cmd = 'insert into members (user_id, group_id) values (%s, %s)'
            values = updated_groups - existing_groups
            sequence = zip([record_id] * len(values), values)
            db.execute_many(cmd, sequence)
    else:
        logger.debug('members unchanged')


def update_group_relationships(record):
    admin = AdminModel(context.site.db)
    admin.update_group_relationships(record)


def send_invitation(user):
    """Send invitation email"""
    password = uuid.uuid4().hex[-10:]
    user.set_password(password)
    site = zoom.system.site
    body = zoom.tools.load_content(
        'welcome', user=user, site=site, password=password)
    subject = 'Welcome - ' + site.name
    zoom.mail.send(user.email, subject, body)
    zoom.alerts.success(f'invitation sent to {user.username}')


def send_password(user, password):
    """Send password reset email"""
    site = zoom.system.site
    body = zoom.tools.load_content(
        'reset', user=user, site=site, password=password)
    subject = 'Password reset - ' + site.name
    zoom.mail.send(user.email, subject, body)


def admin_crud_policy():
    """Authourization policy for Admin app collections
    """
    def _policy(_, user, action):
        """Policy rules for shared collection"""

        def can_crud(user):
            """Return True if user can crud this collection
            """
            return user.is_admin or user.is_member('authorizers')

        actions = {
            'create': can_crud,
            'read': can_crud,
            'update': can_crud,
            'delete': can_crud,
        }

        if action not in actions:
            raise Exception('action missing: {}'.format(action))

        return actions.get(action)(user)
    return _policy


class AdminCollection(zoom.collect.Collection):
    """Admin app Collection"""

    allows = admin_crud_policy()


class GroupsCollection(AdminCollection):

    @property
    def has_many_records(self):
        return False




"""
    icons index
"""

import zoom


class MyView(zoom.View):
    """Index View"""

    def index(self):
        """Index page"""
        zoom.requires('fontawesome4')
        zoom.requires('bootstrap-icons')

        layout = zoom.component.load_component('layout')

        return zoom.page(
            layout.format(
                zoom.tools.load('bi-icons.html'),
                zoom.tools.load('fa-icons.html')
            ),
            title='Icons'
        )

    def about(self):
        """About page"""
        content = '{app.description}'
        return zoom.page(
            content.format(app=zoom.system.request.app),
            title='About {app.title}'.format(app=zoom.system.request.app)
        )


main = zoom.dispatch(MyView)



"""
    admin views
"""

from zoom.mvc import DynamicView
from zoom.component import Component
from model import get_index_metrics


class PanelView(DynamicView):
    subtitle = ''


class IndexPageLayoutView(DynamicView):
    pass


class MetricView(DynamicView):

    def __init__(self, model):
        DynamicView.__init__(self)
        self.name, self.url, self.value = model

    def as_upper(self):
        return self.name.upper()

    @property
    def title(self):
        return 'tooltip'


class MetricsView(DynamicView):

    @property
    def metrics(self):
        return Component(MetricView(m) for m in self.model)


class TestView(DynamicView):
    pass

def index_metrics_view(db):
    return MetricsView(get_index_metrics(db))



"""
    zoom alerts
"""


from zoom.components import compose


def success(message):
    compose(success=message)

def warning(message):
    compose(warning=message)

def error(message):
    compose(error=message)
